# -*- coding: utf-8 -*-
import os
import sys
import time
import base64
import string
import logging
import requests
import traceback
import importlib
import threading
import itertools
from random import choice

TIMEOUT = 600
JPG_FORMATS = ['.JPG', '.JPEG', '.jpg', '.jpeg']
PNG_FORMATS = ['.PNG']
PIPELINE_TOKEN_FOREVER = 'Bearer eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpYXQiOjE1Njg2MjcyMDgsIm5iZiI6MTU2ODYyNzIwOCwianRpIjoiZjc3YmM1N2QtYWNlNS00MjllLTk5Y2EtMzc5ZmZjNmI1ODQzIiwiaWRlbnRpdHkiOnsiY3VycmVudF91c2VyIjp7ImlkIjotMSwidGVuYW50X2lkIjoxLCJjcmVhdGVkX2F0IjpudWxsLCJjcmVhdGVkX2J5IjpudWxsLCJ1cGRhdGVkX2F0IjoiV2VkLCAzMSBKdWwgMjAxOSAwNjowMzozOSBHTVQiLCJ1cGRhdGVkX2J5Ijo1MywibmFtZSI6InBpcGVsaW5lQGludGVsbGlmLmNvbSIsIm5pY2tuYW1lIjoicGlwZWxpbmUiLCJmdWxsbmFtZSI6InBpcGVsaW5lIiwidGVsIjpudWxsLCJlbWFpbCI6InBpcGVsaW5lQGludGVsbGlmLmNvbSIsImFkZHJlc3MiOiIiLCJnZW5kZXIiOjAsInR5cGUiOiJlbWFpbCIsInJlbWFyayI6bnVsbCwiYXZhdGFyIjpudWxsLCJzdGF0dXMiOnRydWUsImlzX3VwZGF0ZV9uYW1lIjpmYWxzZSwiaXNfdmVyaWZ5Ijp0cnVlLCJpc19jb25maXJtIjp0cnVlLCJpc190ZW5hbnRfbWFpbiI6ZmFsc2UsInByb3RvY29sX3ZlcnNpb24iOjEsInJvbGVzIjpbXSwidGVuYW50X25hbWUiOiJcdTZkZjFcdTU3MzNcdTRlOTFcdTU5MjlcdTUyYjFcdTk4ZGVcdTYyODBcdTY3MmZcdTY3MDlcdTk2NTBcdTUxNmNcdTUzZjgifX0sImZyZXNoIjpmYWxzZSwidHlwZSI6ImFjY2VzcyIsInVzZXJfY2xhaW1zIjp7ImN1cnJlbnRfcm9sZXMiOlsiZGVmYXVsdCJdfX0.Hs-19WRuDRIUaXzleg85S_REGC8kjXx1gSYpGsKA2zI'

def assert_true(condition, msg):
    if condition:
        print('{}   √'.format(msg))
    else:
        print('{}   ×'.format(msg))
    return condition

def set_up_log():
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s  %(message)s',
                                  '%Y-%m-%d %H:%M:%S')
    console = logging.StreamHandler(stream=sys.stdout)
    console.setFormatter(formatter)
    console.setLevel(logging.DEBUG)
    logger.addHandler(console)

def generate_secret_key(length=8):
    """Generate secret key"""
    return "".join([choice(string.hexdigits).lower() for i in range(length)])

def get_uniq_id():
    return '%s%s' % (time.strftime('%Y%m%d%H%M%S'), generate_secret_key(8))

def b64encode(text):
    if not isinstance(text, bytes):
        text = str(text).encode('utf-8')
    return base64.b64encode(text).decode('utf-8')

def b64decode(secret):
    return base64.b64decode(secret).decode('utf-8')

def cartesian_join(lbs):
    # 获取多标签的笛卡儿积组合
    # input: [{"交通工具": ["摩托", "飞机"]}, {"材质": ["铁", "合金"]}]
    # output 交通工具.材质, ["摩托.铁", "飞机.铁", "摩托.合金", "飞机.合金"], [[0, 0], [1, 0], [0, 1], [1, 1]]
    lb_keys = []
    lb_values = []
    _lbs = lbs.copy()
    while len(_lbs):
        lb = _lbs.pop(0)
        while not isinstance(lb, list):
            lb_keys.append(list(lb)[0])
            lb = lb.get(list(lb)[0])
        else:
            lb_values.append(lb)

    def cov_index(its, lb_values):
        indexes = []
        for index, it in enumerate(its):
            indexes.append(lb_values[index].index(it))
        return indexes

    label_comps = ['.'.join(item) for item in itertools.product(*lb_values)]
    label_indexes = [cov_index(item, lb_values) for item in itertools.product(*lb_values)]
    return lb_keys, label_comps, label_indexes
    
def convert_labels_multi_2_single(labels):
    # output: [{"交通工具.材质": ["摩托.铁", "飞机.铁", "摩托.合金", "飞机.合金"]}]
    if len(labels.get('label')) > 1:
        main_label = {}
        main_keys = []
        main_values = []
        for label in labels.get('label'):
            main_keys.append(list(label)[0])
        comps_keys, comps, indexes = cartesian_join(labels.get('label'))
        main_label.setdefault('.'.join(comps_keys), comps)
        labels['label'] = main_label

        # 转换数据
        for data in labels.get('data'):
            idx = _.index_of(indexes, data.get('label'))
            data['label'] = [idx]

    return labels

def get_img_fmt_quality(img_path):
    _, img_fmt = os.path.splitext(img_path)
    if img_fmt.upper() in JPG_FORMATS:
        quality = 95
    elif img_fmt.upper() in PNG_FORMATS:
        quality = 9
    else:
        logging.error('this image format not correct %s' % img_path)
        return False, ''
    return quality, img_fmt
    
def get_label(label_type, one_picture, img_width=32, img_height=32):
    if label_type == 'box':
        # 格式参考 https://gluon-cv.mxnet.io/build/examples_datasets/detection_custom.html
        label = [4, 5, img_width, img_height]
        for item in one_picture['boxes']:
            item['box'][0] /= float(img_width)
            item['box'][2] /= float(img_width)
            item['box'][1] /= float(img_height)
            item['box'][3] /= float(img_height)
            one_label = [int(item['class'])] + item['box'] + item.get('label', [])
            label += one_label
        
    elif label_type == 'classification':
        lb = one_picture['label']
        label = lb[0] if isinstance(lb, list) else lb
    elif label_type == 'segmentation':
        label = None
    else:
        label = one_picture['label']

    return label

def try_to_request_async(uri, params, try_num=3, interval=5):
    flag, data = False, None
    headers = {'Authorization': PIPELINE_TOKEN_FOREVER}
    while try_num > 0:
        try:
            export_res = requests.post(uri, json=params, timeout=TIMEOUT, headers=headers)
            export_res = export_res.json()
            if export_res.get('status'):
                export_id = export_res.get('data').get('export_id')
                # 定时查询状态
                start_time = time.time()
                json_path = None
                while time.time() - start_time < 10 * 60:
                    time.sleep(interval)
                    logging.info('{}: try to check export status, export_id {}...'.format(time.time(), export_id))
                    status_res = requests.get(uri, params={'export_id': export_id}, timeout=TIMEOUT, headers=headers)
                    status_res = status_res.json()
                    if status_res.get('status'):
                        json_path = status_res.get('data').get('path')
                        if json_path:
                            flag, data = True, json_path
                            break
                if flag:
                    break
        except Exception as e:
            logging.error(e)
            traceback.print_exc()
            try_num -= 1

        time.sleep(10)

    if flag is False:
        sys.exit(1)
    logging.info('response: {}'.format(data))
    return data, export_res.get('data')

def try_to_request(uri, params, try_num=3):
    flag, data = False, None
    headers = {'Authorization': PIPELINE_TOKEN_FOREVER}
    while try_num > 0:
        try:
            res = requests.post(uri, json=params, timeout=TIMEOUT, headers=headers)
            res = res.json()
            if res.get('status'):
                logging.info('deploy to %s succeed' % uri)
                flag, data = True, res.get('data')
                break
            else:
                msg = res['message']
                logging.info('deploy to %s failed, the reason is %s' % (uri, msg))
                try_num -= 1
        except Exception as e:
            logging.error(e)
            traceback.print_exc()
            try_num -= 1

        time.sleep(10)

    if flag is False:
        sys.exit(1)
    logging.info('response: {}'.format(data))
    return data

def try_to_get_request(uri, params, try_num=3):
    flag, data = False, None
    headers = {'Authorization': PIPELINE_TOKEN_FOREVER}
    while try_num > 0:
        try:
            res = requests.get(uri, params=params, timeout=TIMEOUT, headers=headers)
            res = res.json()
            if res.get('status'):
                logging.info('deploy to %s succeed' % uri)
                flag, data = True, res.get('data')
                break
            else:
                msg = res['message']
                logging.info('deploy to %s failed, the reason is %s' % (uri, msg))
                try_num -= 1
        except Exception as e:
            logging.error(e)
            traceback.print_exc()
            try_num -= 1

        time.sleep(10)

    if flag is False:
        sys.exit(1)
    logging.info('response: {}'.format(data))
    return data

def load_module(pyfile):
    '''动态加载模块

    `pyfile`: py脚本绝对路径, /xxx/script.py
    '''
    pyfile = pyfile.rstrip('.py')
    py_dir = os.path.dirname(pyfile)
    sys.path.append(py_dir)
    prefix = os.path.basename(pyfile)  # 不能含有.py的后缀即不能是文件
    return importlib.import_module(prefix)


# -*- coding: utf-8 -*-
"""aios training
"""
import os
import re
import sys
import logging
import argparse
import mxnet as mx
from aios.utils.db_utils import *
from aios.utils.utils import load_module, assert_true

__all__ = ['load_train_module']

TP_DB_URL = os.getenv('TP_DB_URL')
TENANT_ID = os.getenv('TENANT_ID')
TRAIN_OUTPUT_PATH = os.getenv('TRAIN_OUTPUT_PATH')
CLASSES = os.getenv('CLASSES')

PREFIX = os.getenv('PREFIX')
MODEL_TYPE = os.getenv('MODEL_TYPE')
MODEL_LOAD_EPOCH = os.getenv('MODEL_LOAD_EPOCH')

TRAIN_HYPER_PARAM = os.getenv('TRAIN_HYPER_PARAM')

# 校验：模块加载的时候从环境变量中获取参数
def _check_env():
    logging.info('train:参数校验...')
    # assert TP_DB_URL is not None, 'TP_DB_URL(TP数据库连接字符串)不能为空!'
    # assert TENANT_ID is not None, 'TENANT_ID(企业id)不能为空!'
    # assert TRAIN_OUTPUT_PATH is not None, 'TRAIN_OUTPUT_PATH(训练节点输出模型路径)不能为空!'
    # assert CLASSES is not None, 'CLASSES(类别)不能为空!'
    # 有可能不需要模型
    # assert PREFIX is not None, 'PREFIX(模型)不能为空!'
    # assert MODEL_TYPE is not None, 'MODEL_TYPE(模型类别)不能为空!'
    # assert MODEL_LOAD_EPOCH is not None, 'MODEL_LOAD_EPOCH(模型迭代次数)不能为空!'
    result = []
    result.append(assert_true(TP_DB_URL is not None, 'TP_DB_URL:TP数据库连接字符串'))
    result.append(assert_true(TENANT_ID is not None, 'TENANT_ID:企业id'))
    result.append(assert_true(TRAIN_OUTPUT_PATH is not None, 'TRAIN_OUTPUT_PATH:训练节点输出模型路径'))
    result.append(assert_true(CLASSES is not None, 'CLASSES:类别'))
    return False not in result

def load_train_module(train_script: str):
    '''加载训练脚本模块，训练函数是用训练模块自主调用并输出各项结果到TRAIN_OUTPUT_PATH中

    `Params`:
    1. train_script: 训练脚本绝对路径, 

    `Returns`:
    1. train_module: 以module的形式动态加载的训练脚本, 脚本中如果包含'set_args'函数, 则会执行
    '''
    if not _check_env():
        return

    parser = argparse.ArgumentParser(description="train", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    
    kwargs = {
        'output_path' : TRAIN_OUTPUT_PATH,
        'classes' : CLASSES
    }
    if PREFIX:
        kwargs['prefix'] = PREFIX
        kwargs['model_type'] = MODEL_TYPE
        kwargs['model_load_epoch'] = MODEL_LOAD_EPOCH
    
    if TRAIN_HYPER_PARAM:
        for i in re.findall(r'[--]*([^\s]\S*)\s+([^\s]\S*)', TRAIN_HYPER_PARAM):
            kwargs[i[0]] = i[1]
    parser.set_defaults(**kwargs)
    # 向训练脚本传递参数
    train_module = load_module(train_script)
    if hasattr(train_module, 'set_args'):
        train_module.set_args(parser.parse_args())
    else:
        logging.warning('训练脚本缺少\'set_args\'函数, 无法传入初始化参数!')

    return train_module
    



# -*- coding: utf-8 -*-
"""aios dataset
"""
import os
import sys
import json
import moment
import logging
import datetime
import argparse
import mxnet as mx
import numpy as np
from pytz import timezone
from aios.utils.rec_pack import MultiProcessRecPack, SegmentationPack
from aios.utils.utils import *
from aios.utils.db_utils import *
from aios.utils.other import SyntheticDataIter
from aios.cls.dataset import DatasetClass

__all__ = ['dataset_list', 'load', 'get_dataset']

TP_DB_URL = os.getenv('TP_DB_URL')
DLP_DB_URL = os.getenv('DLP_DB_URL')
ASYNC_EXPORT_URL = os.getenv('ASYNC_EXPORT_URL', 'http://aios-dlp-python/api/v2/dataset_export_bg_async/operation')
DATASET_LIST_URL = os.getenv('DATASET_LIST_URL', 'http://aios-dlp-v1/api/intellif/importData/find?algorithm=classification&pageSize=100000')

FRAMEWORK = os.getenv('FRAMEWORK')
ALGORITHM_TYPE = os.getenv('ALGORITHM_TYPE')
MOUNT_PATH = os.getenv('MOUNT_PATH', '/datakubeflow')
IMAGE_SHAPE = os.getenv('IMAGE_SHAPE', '32,32')
DATASET_OUTPUT_PATH = os.getenv('DATASET_OUTPUT_PATH')
LABEL_FILTER = os.getenv('LABEL_FILTER')

TENANT_ID = os.getenv('TENANT_ID')
USER_ID = os.getenv('USER_ID')
USER_NAME = os.getenv('USER_NAME')

# 校验：模块加载的时候从环境变量中获取参数
def _check_env():
    logging.info('dataset:参数校验...')
    # assert TP_DB_URL is not None, 'TP_DB_URL(TP数据库连接字符串)不能为空!'
    # assert FRAMEWORK is not None, 'FRAMEWORK(数据集打包,框架)不能为空!'
    # assert ALGORITHM_TYPE is not None, 'ALGORITHM_TYPE(算法类型)不能为空!'
    # assert DATASET_OUTPUT_PATH is not None, 'DATASET_OUTPUT_PATH(数据集打包输出路径)不能为空!'
    # assert LABEL_FILTER is not None, 'LABEL_FILTER(数据集标签筛选)不能为空!'
    # assert TENANT_ID is not None, 'TENANT_ID(企业id)不能为空!'
    # assert USER_ID is not None, 'USER_ID(用户id)不能为空!'
    # assert USER_NAME is not None, 'USER_NAME(用户名称)不能为空!'
    result = []
    result.append(assert_true(TP_DB_URL is not None, 'TP_DB_URL:TP数据库连接字符串'))
    result.append(assert_true(FRAMEWORK is not None, 'FRAMEWORK:数据集打包,框架'))
    result.append(assert_true(ALGORITHM_TYPE is not None, 'ALGORITHM_TYPE:算法类型'))
    result.append(assert_true(DATASET_OUTPUT_PATH is not None, 'DATASET_OUTPUT_PATH:数据集打包输出路径'))
    result.append(assert_true(LABEL_FILTER is not None, 'LABEL_FILTER:数据集标签筛选'))
    result.append(assert_true(TENANT_ID is not None, 'TENANT_ID:企业id'))
    result.append(assert_true(USER_ID is not None, 'USER_ID:用户id'))
    result.append(assert_true(USER_NAME is not None, 'USER_NAME:用户名称'))
    return False not in result
        
def _add_data_args(parser):
    data = parser.add_argument_group('Data', 'the input images')
    data.add_argument('--data-train', type=str, help='the training data')
    data.add_argument('--data-train-idx', type=str, default='', help='the index of training data')
    data.add_argument('--data-val', type=str, help='the validation data')
    data.add_argument('--data-val-idx', type=str, default='', help='the index of validation data')
    data.add_argument('--rgb-mean', type=str, default='123.68,116.779,103.939',
                      help='a tuple of size 3 for the mean rgb')
    data.add_argument('--rgb-std', type=str, default='1,1,1',
                      help='a tuple of size 3 for the std rgb')
    data.add_argument('--pad-size', type=int, default=0,
                      help='padding the input image')
    data.add_argument('--fill-value', type=int, default=127,
                      help='Set the padding pixels value to fill_value')
    data.add_argument('--image-shape', type=str, 
                      help='the image shape feed into the network, e.g. (3,224,224)')
    data.add_argument('--num-classes', type=int, help='the number of classes')
    data.add_argument('--num-examples', type=int, help='the number of training examples')
    data.add_argument('--data-nthreads', type=int, default=4,
                      help='number of threads for data decoding')
    data.add_argument('--benchmark', type=int, default=0,
                      help='if 1, then feed the network with synthetic data')
    return data

def _add_data_aug_args(parser):
    aug = parser.add_argument_group(
        'Image augmentations', 'implemented in src/io/image_aug_default.cc')
    aug.add_argument('--random-crop', type=int, default=0,
                     help='if or not randomly crop the image')
    aug.add_argument('--random-mirror', type=int, default=0,
                     help='if or not randomly flip horizontally')
    aug.add_argument('--max-random-h', type=int, default=0,
                     help='max change of hue, whose range is [0, 180]')
    aug.add_argument('--max-random-s', type=int, default=0,
                     help='max change of saturation, whose range is [0, 255]')
    aug.add_argument('--max-random-l', type=int, default=0,
                     help='max change of intensity, whose range is [0, 255]')
    aug.add_argument('--min-random-aspect-ratio', type=float, default=None,
                     help='min value of aspect ratio, whose value is either None or a positive value.')
    aug.add_argument('--max-random-aspect-ratio', type=float, default=0,
                     help='max value of aspect ratio. If min_random_aspect_ratio is None, '
                          'the aspect ratio range is [1-max_random_aspect_ratio, '
                          '1+max_random_aspect_ratio], otherwise it is '
                          '[min_random_aspect_ratio, max_random_aspect_ratio].')
    aug.add_argument('--max-random-rotate-angle', type=int, default=0,
                     help='max angle to rotate, whose range is [0, 360]')
    aug.add_argument('--max-random-shear-ratio', type=float, default=0,
                     help='max ratio to shear, whose range is [0, 1]')
    aug.add_argument('--max-random-scale', type=float, default=1,
                     help='max ratio to scale')
    aug.add_argument('--min-random-scale', type=float, default=1,
                     help='min ratio to scale, should >= img_size/input_shape. '
                          'otherwise use --pad-size')
    aug.add_argument('--max-random-area', type=float, default=1,
                     help='max area to crop in random resized crop, whose range is [0, 1]')
    aug.add_argument('--min-random-area', type=float, default=1,
                     help='min area to crop in random resized crop, whose range is [0, 1]')
    aug.add_argument('--min-crop-size', type=int, default=-1,
                     help='Crop both width and height into a random size in '
                          '[min_crop_size, max_crop_size]')
    aug.add_argument('--max-crop-size', type=int, default=-1,
                     help='Crop both width and height into a random size in '
                          '[min_crop_size, max_crop_size]')
    aug.add_argument('--brightness', type=float, default=0,
                     help='brightness jittering, whose range is [0, 1]')
    aug.add_argument('--contrast', type=float, default=0,
                     help='contrast jittering, whose range is [0, 1]')
    aug.add_argument('--saturation', type=float, default=0,
                     help='saturation jittering, whose range is [0, 1]')
    aug.add_argument('--pca-noise', type=float, default=0,
                     help='pca noise, whose range is [0, 1]')
    aug.add_argument('--random-resized-crop', type=int, default=0,
                     help='whether to use random resized crop')
    return aug

def _add_fit_args(parser):
    """
    parser : argparse.ArgumentParser
    return a parser added with args required by fit
    """
    train = parser.add_argument_group('Training', 'model training')
    train.add_argument('--network', type=str,
                       help='the neural network to use')
    train.add_argument('--num-layers', type=int,
                       help='number of layers in the neural network, \
                             required by some networks such as resnet')
    train.add_argument('--gpus', type=str,
                       help='list of gpus to run, e.g. 0 or 0,2,5. empty means using cpu')
    train.add_argument('--kv-store', type=str, default='device',
                       help='key-value store type')
    train.add_argument('--num-epochs', type=int, default=100,
                       help='max num of epochs')
    train.add_argument('--lr', type=float, default=0.1,
                       help='initial learning rate')
    train.add_argument('--lr-factor', type=float, default=0.1,
                       help='the ratio to reduce lr on each step')
    train.add_argument('--lr-step-epochs', type=str,
                       help='the epochs to reduce the lr, e.g. 30,60')
    train.add_argument('--initializer', type=str, default='default',
                       help='the initializer type')
    train.add_argument('--optimizer', type=str, default='sgd',
                       help='the optimizer type')
    train.add_argument('--mom', type=float, default=0.9,
                       help='momentum for sgd')
    train.add_argument('--wd', type=float, default=0.0001,
                       help='weight decay for sgd')
    train.add_argument('--batch-size', type=int, default=32,
                       help='the batch size')
    train.add_argument('--disp-batches', type=int, default=20,
                       help='show progress for every n batches')
    train.add_argument('--model-prefix', type=str,
                       help='model prefix')
    train.add_argument('--save-period', type=int, default=1, help='params saving period')
    parser.add_argument('--monitor', dest='monitor', type=int, default=0,
                        help='log network parameters every N iters if larger than 0')
    train.add_argument('--load-epoch', type=int,
                       help='load the model on an epoch using the model-load-prefix')
    train.add_argument('--top-k', type=int, default=0,
                       help='report the top-k accuracy. 0 means no report.')
    train.add_argument('--loss', type=str, default='',
                       help='show the cross-entropy or nll loss. ce strands for cross-entropy, nll-loss stands for likelihood loss')
    train.add_argument('--test-io', type=int, default=0,
                       help='1 means test reading speed without training')
    train.add_argument('--dtype', type=str, default='float32',
                       help='precision: float32 or float16')
    train.add_argument('--gc-type', type=str, default='none',
                       help='type of gradient compression to use, \
                             takes `2bit` or `none` for now')
    train.add_argument('--gc-threshold', type=float, default=0.5,
                       help='threshold for 2bit gradient compression')
    # additional parameters for large batch sgd
    train.add_argument('--macrobatch-size', type=int, default=0,
                       help='distributed effective batch size')
    train.add_argument('--warmup-epochs', type=int, default=5,
                       help='the epochs to ramp-up lr to scaled large-batch value')
    train.add_argument('--warmup-strategy', type=str, default='linear',
                       help='the ramping-up strategy for large batch sgd')
    train.add_argument('--profile-worker-suffix', type=str, default='',
                       help='profile workers actions into this file. During distributed training\
                             filename saved will be rank1_ followed by this suffix')
    train.add_argument('--profile-server-suffix', type=str, default='',
                       help='profile server actions into a file with name like rank1_ followed by this suffix \
                             during distributed training')
    train.add_argument('--use-imagenet-data-augmentation', type=int, default=0,
                       help='enable data augmentation of ImageNet data, default disabled')
    return train

def _export_dataset(async_export_url: str, data_id: int, label_filter: str, algorithm_type: str):
    '''从指定API导出数据集

    `label_filter`: 
    >>> [['水果', '苹果'], ['水果', '香蕉']]
    '''
    # has_label: 0没有标签的数据，1是有该标签的数据
    label_filter = json.loads(label_filter)
    params = {
        'data_id': data_id,
        'has_label': 1,
        'clazz': label_filter,
        'algorithm_type': algorithm_type,
        'source': 'pipeline'
    }
    logging.info('start export dataset...')
    return try_to_request_async(async_export_url, params)

def _repack(data_id, is_none_tp_dataset):
    '''重新打包最新的数据集
    '''
    try:
        parent_dir = os.path.dirname(DATASET_OUTPUT_PATH)
        try:
            # 存在两个节点同时运行,其中一个会报错
            if not os.path.exists(parent_dir):
                os.makedirs(parent_dir)
        except Exception as e:
            logging.error(e)

        label_json, export_res = _export_dataset(ASYNC_EXPORT_URL, data_id, LABEL_FILTER, ALGORITHM_TYPE)
        absolute_label_json = os.path.join(MOUNT_PATH.rstrip('/'), label_json.lstrip('/')).replace('/', os.path.sep)
        if not os.path.exists(absolute_label_json):
            logging.error('File not exist, {}'.format(absolute_label_json))
            sys.exit(1)
        # 导出的json文件
        labels = json.load(open(absolute_label_json, mode='rb'))

        logging.info('framework:{}, type_:{}'.format(FRAMEWORK, ALGORITHM_TYPE))
        logging.info('start pack...')
        if FRAMEWORK == 'mxnet':
            if ALGORITHM_TYPE == 'segmentation':
                SegmentationPack(mount_path=MOUNT_PATH).create_segmentation_txt(DATASET_OUTPUT_PATH, labels)
            elif ALGORITHM_TYPE == 'classification':
                # 转换多标签，组合成单标签
                labels = convert_labels_multi_2_single(labels)
                MultiProcessRecPack(mount_path=MOUNT_PATH, image_shape=IMAGE_SHAPE).mxnet_multiprocess_handle(DATASET_OUTPUT_PATH, labels)
            else:
                MultiProcessRecPack(mount_path=MOUNT_PATH, image_shape=IMAGE_SHAPE).mxnet_multiprocess_handle(DATASET_OUTPUT_PATH, labels)
            # 打包完成之后，记录数据集信息，更新缓存
            time_now = moment.date(datetime.datetime.now(timezone('UTC'))).format("YYYY-MM-DD HH:mm:ss")
            # sql语法转化
            tag = 'NULL' if export_res.get('uuid') is None else "'{}'".format(export_res.get('uuid'))
            pack_pos = 'NULL' if DATASET_OUTPUT_PATH is None else "'{}'".format(DATASET_OUTPUT_PATH)

            if is_none_tp_dataset is False:
                name = export_res.get('dataset_name')
                tenant_id = export_res.get('tenant_id')
                sql = "insert into dataset_model(name, data_id, tag, tenant_id, pack_pos, created_by, created_at, owner) values('{}', {}, {}, {}, {}, {}, '{}', '{}');".format(name, data_id, tag, tenant_id, pack_pos, USER_ID, time_now, USER_NAME)
            else:
                sql = "update dataset_model set tag={}, pack_pos={}, updated_by={}, updated_at='{}' where data_id={};".format(tag, pack_pos, USER_ID, time_now, data_id)
            
            logging.info('更新TP数据库数据集记录...')
            execute(TP_DB_URL, sql)
            # red file
            return DATASET_OUTPUT_PATH
        else:
            logging.info('暂不支持的框架, {}'.format(FRAMEWORK))
            sys.exit(1)
    except Exception as err:
        logging.error('repack error, {}'.format(err))

def get_dataset(dataset_obj):
    '''获取数据集，从历史缓存获取/重新打包

    `Returns`:
    1. rec file
    '''

    # 检测数据集是否有变动
    tp_sql = 'select tag, pack_pos from dataset_model where data_id = {};'.format(dataset_obj.data_id)
    tp_dataset = query_one(TP_DB_URL, tp_sql)

    logging.info('dlp数据库中uuid:{}, tp数据库中tag:{}'.format(dataset_obj.tag, tp_dataset[0]))
    if tp_dataset is None or dataset_obj.tag != tp_dataset[0]:
        # 拉取最新数据集打包
        is_none_tp_dataset = bool(tp_dataset)
        rec_file = _repack(dataset_obj.data_id, is_none_tp_dataset)
        logging.info('拉取最新数据集打包')
    else:
        # 获取历史数据集recfile
        rec_file = tp_dataset.pack_pos
        logging.info('获取历史数据集recfile')
    
    return rec_file

def load(dataset_obj: DatasetClass, predicate=None):
    '''
    `Params`:  
        dataset_obj (DatasetClass): (cls.dataset).
        predicate (mixed): Predicate applied per iteration.

    `Return`: 
        Iter: Filtered rec iter.
    '''
    if not _check_env():
        return

    parser = argparse.ArgumentParser(description="train imagenet-1k",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    _add_fit_args(parser)
    _add_data_args(parser)
    _add_data_aug_args(parser)

    parser.set_defaults(
        # network
        network          = 'resnet',
        num_layers       = 50,
        # data
        num_classes      = 1000,
        num_examples     = 1281167,
        image_shape      = '3,32,32',
        min_random_scale = 1, # if input image has min size k, suggest to use
                              # 256.0/x, e.g. 0.533 for 480
        # train
        num_epochs       = 80,
        lr_step_epochs   = '30,60',
        dtype            = 'float32'
    )

    args = parser.parse_args()
    # 加载数据集
    data_train = get_dataset(dataset_obj)
    assert os.path.exists(data_train), '找不到rec文件, {}'.format(data_train)

    image_shape = (3,) + tuple([int(l) for l in IMAGE_SHAPE.split(',')])
    if 'benchmark' in args and args.benchmark:
        data_shape = (args.batch_size,) + image_shape
        train = SyntheticDataIter(args.num_classes, data_shape,
                args.num_examples / args.batch_size, np.float32)
        return (train, None)

    # rec file
    train = mx.io.ImageRecordIter(
        path_imgrec=data_train,
        path_imgidx='',
        label_width=1,
        preprocess_threads=args.data_nthreads,
        data_name='data',
        label_name='softmax_label',
        shuffle=True,
        data_shape=image_shape,
        batch_size=args.batch_size,
    )
    return train

def dataset_list():
    '''数据集列表函数，获取当前企业下所有数据集记录的列表
    '''
    _check_env()

    datasets = try_to_get_request(DATASET_LIST_URL, params={'algorithm': ALGORITHM_TYPE, 'pageSize': 100000})
    return [convert_dataset(dataset) for dataset in datasets]



# -*- coding: utf-8 -*-
"""aios models
"""
import os
import sys
import logging
import mxnet as mx
from aios.utils.db_utils import *
from aios.utils.utils import load_module
from aios.cls.model import ModelClass

__all__ = ['model_list', 'load']

TP_DB_URL = os.getenv('TP_DB_URL')
TENANT_ID = os.getenv('TENANT_ID')

def _check_env():
    # 校验：模块加载的时候从环境变量中获取参数
    logging.info('model:参数校验...')
    assert TP_DB_URL is not None, 'TP_DB_URL(TP数据库连接字符串)不能为空!'
    assert TENANT_ID is not None, 'TENANT_ID(企业id)不能为空!'

def _get_symbol(args):
    """
    :param args:
    :return:
    """
    args.depth = 50
    if args.format == 'python':
        model = load_module(args.prefix)

        # depth should be one of 110, 164, 1001,...,which is should fit (args.depth-2)%9 == 0
        if((args.depth-2)%9 == 0 and args.depth >= 164):
            per_unit = [(args.depth-2)/9]
            filter_list = [16, 64, 128, 256]
            bottle_neck = True
        elif((args.depth-2)%6 == 0 and args.depth < 164):
            per_unit = [(args.depth-2)/6]
            filter_list = [16, 16, 32, 64]
            bottle_neck = False
        else:
            raise ValueError("no experiments done on detph {}, you can do it youself".format(args.depth))
        units = per_unit*3
        symbol = model.resnet(units=units, num_stage=3, filter_list=filter_list, num_class=args.num_classes,
                        data_type="cifar10", bottle_neck=bottle_neck, bn_mom=args.bn_mom, workspace=args.workspace,
                        memonger=args.memonger)
        print(symbol)
        arg_params = None
        aux_params = None
    else:
        # 接着训练
        #symbol, _arg_params, aux_params = mx.model.load_checkpoint(args.prefix, args.model_load_epoch)

        # fine-tune, resnet.py一起上传，用reset.py获取模型结构，用与训练模型初始化参数
        from importlib import import_module
        try:
            model = import_module('resnet')
        except ModuleNotFoundError as err:
            raise Exception('找不到resnet模块，请检查是否已经上传，{}'.format(err))
        # depth should be one of 110, 164, 1001,...,which is should fit (args.depth-2)%9 == 0
        if((args.depth-2)%9 == 0 and args.depth >= 164):
            per_unit = [(args.depth-2)/9]
            filter_list = [16, 64, 128, 256]
            bottle_neck = True
        elif((args.depth-2)%6 == 0 and args.depth < 164):
            per_unit = [(args.depth-2)/6]
            filter_list = [16, 16, 32, 64]
            bottle_neck = False
        else:
            raise ValueError("no experiments done on detph {}, you can do it youself".format(args.depth))
        units = per_unit*3
        symbol = model.resnet(units=units, num_stage=3, filter_list=filter_list, num_class=args.num_classes,
                        data_type="cifar10", bottle_neck=bottle_neck, bn_mom=args.bn_mom, workspace=args.workspace,
                        memonger=args.memonger)

        # args.model_load_epoch是resume训练的时候用到，预训练的时候最好换个参数名
        __, _arg_params, aux_params = mx.model.load_checkpoint(args.prefix, args.model_load_epoch)
        arg_params = {}
        for key in _arg_params.keys():
            if key == 'fc1_bias' or key == 'fc1_weight':
                continue
            arg_params[key] = _arg_params[key]
    return symbol, arg_params, aux_params

def model_list():
    '''模型列表函数，获取当前企业下所有模型记录的列表
    '''
    _check_env()

    sql = 'select id, name, tenant_id, framework, type, path, origin, info, size, algorithm_type, version, format from model_model where tenant_id={} order by id asc;'.format(TENANT_ID)
    # convert to object
    models = query_many(TP_DB_URL, sql)
    return [convert_model(model) for model in models]

def load(model_obj: ModelClass):
    '''模型加载函数，模型分为多种格式：
    1. py文件
    2. params/json文件
    '''
    _check_env()
    
    assert model_obj.path is not None, 'model的path属性不能为空'
    assert os.path.exists(model_obj.path), 'model目录不存在'

    if model_obj.format == 'python':
        entry_module = None
        pyfiles = []
        for root, dirs, files in os.walk(model_obj.path):
            sys.path.append(root)
            for file in filter(lambda x: x.endswith('.py'), files):
                pyfiles.append(file)
        if len(pyfiles) >= 2 and 'main.py' not in pyfiles:
            raise Exception('错误的脚本，文件目录下找不到入口文件 main.py')
        entry_file = 'main.py' if 'main.py' in pyfiles else pyfiles[0]
        logging.info('加载python文件, {}'.format(entry_file))
        return load_module(entry_file[:-3])
    else:
        if model_obj.framework == "mxnet":
            params_file, symbol_json_file = None, None
            for it in os.listdir(model_obj.path):
                path = os.path.join(model_obj.path, it)
                if it.split(".")[-1] == "params":
                    params_file = path
                elif it.split(".")[-1] == "json":
                    symbol_json_file = path
            if not params_file or not symbol_json_file:
                raise Exception("params file or json file not found")

            model_obj.prefix = symbol_json_file[0: -len('-symbol.json')]
            model_obj.model_load_epoch = params_file[-11:-7]
            return _get_symbol(model_obj)
        else:
            raise Exception('不支持的framework, {}'.format(model_obj.framework))



# -*- coding: utf-8 -*-
"""aios evaluate
"""
import os
import sys
import json
import logging
from aios.utils.utils import PIPELINE_TOKEN_FOREVER, try_to_request

__all__ = ['deploy']

TENANT_ID = os.getenv('TENANT_ID')
CLASSES = os.getenv('CLASSES')
IMAGE_SHAPE = os.getenv('IMAGE_SHAPE', '32,32')
CP_SVC_URI = os.getenv('CP_SVC_URI')
MODEL_INFO = os.getenv('MODEL_INFO')
POST_DATA = os.getenv('POST_DATA')
MODEL_TYPE = os.getenv('MODEL_TYPE')

# 校验：模块加载的时候从环境变量中获取参数
def _check_env():
    logging.info('eval:参数校验...')
    assert TENANT_ID is not None, 'TENANT_ID(企业id)不能为空!'
    assert CLASSES is not None, 'CLASSES(类别)不能为空!'
    assert CP_SVC_URI is not None, 'CP_SVC_URI(CP部署服务url)不能为空!'
    assert POST_DATA is not None, 'POST_DATA(CP部署服务url参数)不能为空!'
    assert MODEL_INFO is not None, 'MODEL_INFO(模型目录)不能为空!'
    assert MODEL_TYPE is not None, 'MODEL_TYPE(模型类型)不能为空!'

def _get_model_real_path(model_folder):
    model_prefix = None
    load_epoch = None
    for it in os.listdir(model_folder):
        if it.endswith("params"):
            load_epoch = int(it[-11:-7])
            model_prefix = os.path.join(model_folder, it[0:-len('-0000.params')])
            break
    if not model_prefix:
        return False, "params file or json file not found", ""
    return True, model_prefix, load_epoch

def deploy():
    """参数传递考虑了两种方式，一是arg_parse, 另外是**kwargs
    """
    try:
        logging.info('start deploy action ....')
        
        data = json.loads(POST_DATA)
        model_prefix = None
        load_epoch = 0
        logging.info('post data is :')
        logging.info(data)
        if MODEL_TYPE in ['evaluate', 'active_learning']:
            # eval之后才deploy，判断evaluate的结果
            if not os.path.exists(MODEL_INFO):
                logging.info('record file not exist')
                sys.exit(1)

            with open(MODEL_INFO) as fo:
                ret = fo.readlines()
                if len(ret) != 3:
                    logging.info('evaluate result is not correct')
                    sys.exit(1)
                if ret[2].strip() == '1':  # 是否更新的标志位
                    model_folder = ret[0].strip()  # 换行符的处理
                    flag, model_prefix, load_epoch = _get_model_real_path(model_folder)
        else:
            # 直接进入到deploy
            model_prefix = MODEL_INFO

        if model_prefix:
            logging.info('model prefix is : %s' % model_prefix)
            cmd = 'python3 /app/tptool/tp_tool_deploy.py --script-path {} --mparser-model-prefix {} --mparser-load-epoch {} --classes {} --image-shape {}'.\
                format(data['containers'][0]['script_path'], model_prefix, load_epoch, CLASSES, IMAGE_SHAPE)   # @TODO --classes如果传的json字符串会解析出错
            if data['containers'][0]['sub_param']:
                cmd += ' --sub-param={}'.format(data['containers'][0]['sub_param'])
            if POST_DATA:
                cmd += ' --config %s' % POST_DATA
            logging.info('cmd is: %s' % cmd)
            data['containers'][0]['cmd'] = cmd
            try_to_request(CP_SVC_URI, data, PIPELINE_TOKEN_FOREVER)

        else:
            logging.error('get model prefix failed')

    except Exception as err:
        logging.error('occured: {}'.format(str(err)))
        import traceback
        traceback.print_exc()
        sys.exit(1)


# -*- coding: utf-8 -*-
import mxnet as mx
import numpy as np
from mxnet.io import DataBatch, DataIter

class SyntheticDataIter(DataIter):
    def __init__(self, num_classes, data_shape, max_iter, dtype):
        self.batch_size = data_shape[0]
        self.cur_iter = 0
        self.max_iter = max_iter
        self.dtype = dtype
        label = np.random.randint(0, num_classes, [self.batch_size,])
        data = np.random.uniform(-1, 1, data_shape)
        self.data = mx.nd.array(data, dtype=self.dtype, ctx=mx.Context('cpu_pinned', 0))
        self.label = mx.nd.array(label, dtype=self.dtype, ctx=mx.Context('cpu_pinned', 0))
    def __iter__(self):
        return self
    @property
    def provide_data(self):
        return [mx.io.DataDesc('data', self.data.shape, self.dtype)]
    @property
    def provide_label(self):
        return [mx.io.DataDesc('softmax_label', (self.batch_size,), self.dtype)]
    def next(self):
        self.cur_iter += 1
        if self.cur_iter <= self.max_iter:
            return DataBatch(data=(self.data,),
                             label=(self.label,),
                             pad=0,
                             index=None,
                             provide_data=self.provide_data,
                             provide_label=self.provide_label)
        else:
            raise StopIteration
    def __next__(self):
        return self.next()
    def reset(self):
        self.cur_iter = 0



# -*- coding: utf-8 -*-
from sqlalchemy.engine import create_engine


def get_engine(db_conn):
    return create_engine(db_conn)

def query_one(db_conn, sql):
    item = None
    with get_engine(db_conn).connect() as conn:
        item = conn.execute(sql).fetchone()
    return item

def query_many(db_conn, sql):
    items = None
    with get_engine(db_conn).connect() as conn:
        items = conn.execute(sql).fetchall()
    return items

def execute(db_conn, sql):
    items = None
    with get_engine(db_conn).connect() as conn:
        items = conn.execute(sql)
    return items

def convert_model(m_tuple: tuple):
    from aios.cls.model import ModelClass
    return ModelClass(id=m_tuple[0],
                      name=m_tuple[1],
                      tenant_id=m_tuple[2],
                      framework=m_tuple[3],
                      type=m_tuple[4],
                      path=m_tuple[5],
                      origin=m_tuple[6],
                      info=m_tuple[7],
                      size=m_tuple[8],
                      algorithm_type=m_tuple[9],
                      version=m_tuple[10],
                      format=m_tuple[11])


def convert_dataset(d_dict: dict):
    from aios.cls.dataset import DatasetClass
    return DatasetClass(id=d_dict.get('id'),
                        name=d_dict.get('name'),
                        tag=d_dict.get('uuid'),
                        data_id=d_dict.get('id'),
                        tenant_id=d_dict.get('tenantId'))


# -*- coding: utf-8 -*-
"""
    image classification training script, use cifa10 rec file, datashape is 3*32*32
    使用预训练模型fine-tune
"""

import logging
import mxnet as mx
import argparse
import os
import sys
import importlib

os.environ['MXNET_CUDNN_AUTOTUNE_DEFAULT'] = '0'

logging.basicConfig(format='%(asctime)s [%(levelname)s] %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',level=logging.INFO)
logging.info('start to train')


def log_train_metric(period, writer, auto_reset=False):
    """Callback to log the training evaluation result every period.
    inlucding tensorboard log and progress log
    period : int, The number of batch to log the training evaluation metric.
    auto_reset : bool, Reset the metric after each log.

    Returns
    -------
    callback : function
        The callback function that can be passed as iter_epoch_callback to fit.
    """
    def _callback(param):
        """The checkpoint function."""
        if param.nbatch % period == 0 and param.eval_metric is not None:
            name_value = param.eval_metric.get_name_value()
            for name, value in name_value:
                writer.add_scalar(name, value)
                logging.info('Iter[%d] Batch[%d] Train-%s=%f',
                             param.epoch, param.nbatch, name, value)
            if auto_reset:
                param.eval_metric.reset()

    return _callback

def start_train(args):
    """
    :param train_path:
    :param val_path:
    :param prefix: model preffix, only support symbol file,
    :param batch_size:
    :param epoch:
    :return:
    """

    train_path = args.train_data
    val_path = args.val_data
    batch_size = args.batch_size
    epoch = args.num_epochs
    
    [height, width] = args.image_shape.split(',')
    image_shape_width = int(width)
    image_shape_height = int(height)

    param_save_path = _get_model_save_path(args)

    logging.info('start_train')
    logging.info('params save path is %s' % param_save_path)
    logging.info('ImageRecordIter.data_shape is {}, {}'.format(image_shape_height, image_shape_width))
    devs = mx.cpu() if args.gpus is None or args.gpus == "" else [
        mx.gpu(int(i)) for i in args.gpus.split(',')]

    sym, arg_params, aux_params = _get_symbol(args)

    # load train data and test data
    train_data = mx.io.ImageRecordIter(
        path_imgrec=train_path,
        path_imgidx='',
        label_width=1,
        preprocess_threads=4,
        data_name='data',
        label_name='softmax_label',
        shuffle=True,
        data_shape=(3, image_shape_height, image_shape_width),
        batch_size=batch_size,
    )
    if val_path:
        val_data = mx.io.ImageRecordIter(
            path_imgrec=val_path,
            path_imgidx='',
            label_width=1,
            preprocess_threads=4,
            data_name='data',
            label_name='softmax_label',
            shuffle=True,
            data_shape=(3, image_shape_height, image_shape_width),
            batch_size=batch_size,
        )
    else:
        val_data = None

    mod = mx.mod.Module(
        symbol=sym, context=devs, label_names=(
            'softmax_label',), data_names=(
            'data',))
    data_shapes = train_data.provide_data

    mod.bind(
        for_training=True,
        data_shapes=data_shapes,
        label_shapes=train_data.provide_label)
    if arg_params is not None:
        mod.set_params(arg_params, aux_params, allow_missing=True)
    kv = mx.kv.create(args.kv_store)
    mod.fit(train_data,
            eval_data=val_data,
            num_epoch=epoch,
            kvstore=kv,
            epoch_end_callback=mx.callback.do_checkpoint(param_save_path),
            optimizer='sgd',
            optimizer_params={
                'learning_rate': 0.1,
                'wd': 0.0001,
                'momentum': 0.9,
                'multi_precision': True},
            eval_metric='acc',
            begin_epoch=0,
            )
    logging.info('finish train')

def _get_symbol(args):
    """
    :param args:
    :return:
    """
    args.depth = 50
    if args.model_type == 'python':
        py_dir = os.path.dirname(args.prefix)
        sys.path.append(py_dir)
        prefix = os.path.basename(args.prefix)  # 不能含有.py的后缀即不能是文件
        model = importlib.import_module(prefix)

        # depth should be one of 110, 164, 1001,...,which is should fit (args.depth-2)%9 == 0
        if ((args.depth - 2) % 9 == 0 and args.depth >= 164):
            per_unit = [(args.depth - 2) / 9]
            filter_list = [16, 64, 128, 256]
            bottle_neck = True
        elif ((args.depth - 2) % 6 == 0 and args.depth < 164):
            per_unit = [(args.depth - 2) / 6]
            filter_list = [16, 16, 32, 64]
            bottle_neck = False
        else:
            raise ValueError("no experiments done on detph {}, you can do it youself".format(args.depth))
        units = per_unit*3
        symbol = model.resnet(units=units, num_stage=3, filter_list=filter_list, num_class=args.num_classes,
                        data_type="cifar10", bottle_neck=bottle_neck, bn_mom=args.bn_mom, workspace=args.workspace,
                        memonger=args.memonger)
        print(symbol)
        arg_params = None
        aux_params = None
    else:
        # 接着训练
        #symbol, _arg_params, aux_params = mx.model.load_checkpoint(args.prefix, args.model_load_epoch)

        # fine-tune, resnet.py一起上传，用reset.py获取模型结构，用与训练模型初始化参数
        from importlib import import_module
        model = import_module('resnet')
        # depth should be one of 110, 164, 1001,...,which is should fit (args.depth-2)%9 == 0
        if ((args.depth - 2) % 9 == 0 and args.depth >= 164):
            per_unit = [(args.depth - 2) / 9]
            filter_list = [16, 64, 128, 256]
            bottle_neck = True
        elif ((args.depth - 2) % 6 == 0 and args.depth < 164):
            per_unit = [(args.depth - 2) / 6]
            filter_list = [16, 16, 32, 64]
            bottle_neck = False
        else:
            raise ValueError("no experiments done on detph {}, you can do it youself".format(args.depth))
        units = per_unit*3
        symbol = model.resnet(units=units, num_stage=3, filter_list=filter_list, num_class=args.num_classes,
                        data_type="cifar10", bottle_neck=bottle_neck, bn_mom=args.bn_mom, workspace=args.workspace,
                        memonger=args.memonger)

        # args.model_load_epoch是resume训练的时候用到，预训练的时候最好换个参数名
        _, _arg_params, aux_params = mx.model.load_checkpoint(args.prefix, args.model_load_epoch)
        arg_params = {}
        for key in _arg_params.keys():
            if key == 'fc1_bias' or key == 'fc1_weight':
                continue
            arg_params[key] = _arg_params[key]
    return symbol, arg_params, aux_params

def _get_model_save_path(args):
    base_name = os.path.basename(args.output_path)
    param_save_path = os.path.join(args.output_path, base_name)
    if not os.path.exists(args.output_path):
        os.makedirs(args.output_path)

    return param_save_path

def _get_data_count(test_data):
    data_count = 0
    rec_reader = mx.recordio.MXRecordIO(test_data, 'r')
    while True:
        if not rec_reader.read():
            break
        data_count += 1
    rec_reader.close()

    return data_count

def _init_args():
    parser = argparse.ArgumentParser(description='train a model for image classification')
    parser.add_argument('--prefix', default='resnet', type=str, help='path to checkpoint prefix')
    parser.add_argument('--train-data', type=str, default="", help='training dataset')
    parser.add_argument('--val-data', type=str, default="", help='validation dataset')
    parser.add_argument('--model-load-epoch', type=int, default=0, help='resume model epoch')
    parser.add_argument('--model-type', type=str, default="python", help='json or py')
    parser.add_argument('--classes', type=str, help='label classes')
    parser.add_argument('--output-path', type=str, default="./models/fruit", help='')

    parser.add_argument('--image-shape', type=str, default='32,32', help='图片压缩尺寸')
    parser.add_argument('--batch-size', type=int, default=32, help='training batch size per device (CPU/GPU).')
    parser.add_argument('--optimizer', type=str, default='sgd', help='')
    parser.add_argument('--weight-decay', type=float, default=0.001, help='')
    parser.add_argument('--kv-store', type=str, default='local', help='key-value store type')
    parser.add_argument('--lr', type=float, default=0.01, help='learning rate. default is 0.1.')
    parser.add_argument('--lr-factor', type=float, default=1, help='')
    parser.add_argument('--lr-step-epochs', type=int, default=1, help='')
    parser.add_argument('--gpus', type=str, default=None, help="the gpu id str used for train, '0,1,2'")
    parser.add_argument('--num-epochs', type=int, default=10,help='number of training epochs.')
    parser.add_argument('--shuffle', action='store_true',help='default is false.')

    # optional
    parser.add_argument('--mom', type=float, default=0.9, help='momentum for sgd')
    parser.add_argument('--bn-mom', type=float, default=0.9, help='momentum for batch normlization')
    parser.add_argument('--workspace', type=int, default=512, help='memory space size(MB) used in convolution, if xpu '
                        ' memory is oom, then you can try smaller vale, such as --workspace 256')
    parser.add_argument('--depth', type=int, default=50, help='the depth of resnet')
    parser.add_argument('--num-examples', type=int, default=1281167, help='the number of training examples')
    parser.add_argument('--memonger', action='store_true', default=False,
                        help='true means using memonger to save momory, https://github.com/dmlc/mxnet-memonger')

    args = parser.parse_args()

    args.data_type = "cifar10"
    args.aug_level = 1  # use only random crop and random mirror
    
    # ===========================示例参数===========================
    args.train_data = '/datakubeflow/dev/tp/1/temp/run_pipeline/20190710003110a94d5dda/dataset/1021-a25f4d10-b7af-f097-ae37-c8fa6c207629.rec'
    args.val_data = '/datakubeflow/dev/tp/1/temp/run_pipeline/20190710003110a94d5dda/dataset/1021-daea0c4b-e2d2-76e9-e8a2-941abc067f13.rec'
    args.output_path = '/datakubeflow/dev/tp/1/temp/run_pipeline/20190710003110a94d5dda/model/12-c18f'
    args.classes = '"Apple Braeburn,Apple Crimson Snow,Apple Golden 1"'
    args.num_classes = len(args.classes.split(','))
    # fruit-resnet.py文件
    args.prefix = '/datakubeflow/dev/tp/3/6d219f54e85e8b6784847d13bf7a/fruit-resnet'
    args.model_type = 'python'
    args.model_load_epoch = 0
    # gpu序列, 0,1,2...
    args.gpus = '0'
    # 计算训练集数据量, 可以支持少量数据的调试和批量数据的作业
    # 可以固定指定大小
    args.batch_size = min(args.batch_size, _get_data_count(args.train_data))
    
    logging.info(args)
    return args

if __name__ == '__main__':
    args = _init_args()
    start_train(args)



# -*- coding: utf-8 -*-
"""aios evaluate
"""
import os
import re
import sys
import json
import shutil
import logging
import datetime
import argparse
import mxnet as mx
from aios.utils.db_utils import *
from aios.utils.utils import load_module, b64decode, b64encode, get_uniq_id

__all__ = ['evaluate_models']

TP_DB_URL = os.getenv('TP_DB_URL')
TENANT_ID = os.getenv('TENANT_ID')
FRAMEWORK = os.getenv('FRAMEWORK')
VAL_DATA = os.getenv('VAL_DATA')
EVAL_OUTPUT_PATH = os.getenv('EVAL_OUTPUT_PATH')
CLASSES = os.getenv('CLASSES')
ALGORITHM_TYPE = os.getenv('ALGORITHM_TYPE')
IMAGE_SHAPE = os.getenv('IMAGE_SHAPE', '32,32')

MODEL_ROOT = os.getenv('MODEL_ROOT')

EVAL_MODEL = os.getenv('EVAL_MODEL')
EVAL_SCRIPT = os.getenv('EVAL_SCRIPT')

GPUS = os.getenv('GPUS')

TENANT_ID = os.getenv('TENANT_ID')
USER_ID = os.getenv('USER_ID')
OWNER = os.getenv('OWNER')
ITER = os.getenv('ITER')

# 校验：模块加载的时候从环境变量中获取参数
def _check_env():
    logging.info('eval:参数校验...')
    assert TP_DB_URL is not None, 'TP_DB_URL(TP数据库连接字符串)不能为空!'
    assert TENANT_ID is not None, 'TENANT_ID(企业id)不能为空!'
    assert USER_ID is not None, 'USER_ID(用户id)不能为空!'
    assert OWNER is not None, 'OWNER(创建者)不能为空!'
    assert FRAMEWORK is not None, 'FRAMEWORK(框架)不能为空!'
    assert EVAL_OUTPUT_PATH is not None, 'EVAL_OUTPUT_PATH(训练节点输出模型路径)不能为空!'
    assert CLASSES is not None, 'CLASSES(类别)不能为空!'
    assert ALGORITHM_TYPE is not None, 'ALGORITHM_TYPE(算法类型)不能为空!'
    assert MODEL_ROOT is not None, 'MODEL_ROOT(模型目录)不能为空!'
    assert EVAL_MODEL is not None, 'EVAL_MODEL(待评估模型目录)不能为空!'
    assert EVAL_SCRIPT is not None, 'EVAL_SCRIPT(评估脚本)不能为空!'
    assert ITER is not None, 'ITER(迭代次数)不能为空!'

def _load_evaluate(function):
    """Get eavluate function from evaluation function file
    """
    eval_module = load_module(function.rstrip('.py'))
    return eval_module.evaluate

def _get_model_prefix_epoch(model_path):
    """Get all models in model path
    """
    epochs = []
    model_prefix = None
    for it in os.listdir(model_path):
        path = os.path.join(model_path, it)
        prefix, ext = os.path.splitext(path)
        if ext == ".params":
            epoch = prefix[-4:]
            if not epoch.isdigit():
                continue
            epoch = int(epoch)
            epochs.append(epoch)
            model_prefix = prefix[0:-len('-0000')]
        elif ext == ".json":
            model_prefix = prefix[0:len(prefix)-7]
    
    return model_prefix, epochs

def _save_result(model_prefix, flag, output_path, load_epoch=0):
    """Compare with previous iteration and write the result file
       保存到临时文件，供deploy和inference使用
       文件内容： 第一行,模型路径
                 第二行, load_epoch
                 第三行,0/1, 1表示比上次模型更优
    """
    result = model_prefix + "\n" + str(load_epoch) + "\n" + str(flag)
    file_path = output_path
    try:
        dir = os.path.dirname(file_path)
        if not os.path.exists(dir):
            os.mkdir(dir)
        # 所在目录要存在
        with open(file_path, "w") as f:  # 多此迭代如果目录相同会直接覆盖之前的结果
            f.write(result)
    except Exception as e:
        logging.error(e)
        return False, 'save result failed'
    return True, "Ok"

def _save_model(model_prefix, eval_value, save_epoch):
    """Save current best model to fs and db
        分三种情况py脚本第一次/预训练模型第一次/非第一次训练, 返回历史最优模型的model_prefix,load_epoch都是0
        确认现有的模型存储的接口，如果方便调用直接调用存储接口
        模型存储接口需要jwt认证；
    """
    logging.info('start to add new model to system %s' % model_prefix)
    # Save to fs, 每个json和params单独文件夹，文件夹名和json的前缀相同
    name = '_'.join([os.path.basename(model_prefix), str(ITER), get_uniq_id()])
    save_path = os.path.join(MODEL_ROOT, name)  # 文件夹目录
    if not os.path.exists(save_path):
        os.makedirs(save_path)

    new_model_prefix = os.path.join(save_path, name)
    try:
        symbol_path = model_prefix + "-symbol.json"
        params_path = model_prefix + "-%04d.params" % save_epoch
        if os.path.exists(symbol_path):
            shutil.copy(symbol_path, "{}-symbol.json".format(new_model_prefix))
        # 如果要保留epoch信息，将0000修改为%04d
        shutil.copy(params_path, "{}-0000.params".format(new_model_prefix))
    except Exception as exception:
        logging.error(exception)
        return False, "Save model file failed"
    
    # Save to db
    origin = ""
    size_byte = os.path.getsize(params_path)
    if size_byte < 1024:
        size = str(size_byte) + 'B'
    elif size_byte < 1024 * 1024:
        size = '%.3f' % (size_byte / 1024) + 'KB'
    else:
        size = '%.3f' % (size_byte / (1024 * 1024)) + 'MB'
    info = {
        "iter": ITER,
        "eval_value": eval_value
    }
    try:
        path = os.path.dirname(new_model_prefix)
        sql = "INSERT INTO model_model( name, created_at, owner, created_by, tenant_id, format, framework," \
                  " type, path, origin, size, info) values('{}', '{}','{}','{}', '{}', '{}', '{}', '{}', '{}', '{}', " \
                  "'{}', '{}')".format(name, datetime.datetime.utcnow(), OWNER, USER_ID, TENANT_ID, format, FRAMEWORK,
                                       ALGORITHM_TYPE, path, origin, size, json.dumps(info))
        logging.info(sql)
        execute(TP_DB_URL, sql)
    except Exception as e:
        logging.error(e)
        return False, 'inser database error'

    # 返回model_prefix即数据库中的path,无文件后缀
    logging.info('add new model to system succeed')
    return True, path

def evaluate_models():
    """Eval all models generated by one training task, and save the best model
       评估之后新的模型命名后缀是0000.params,上传的预训练好的模型不一定是0000
    """
    try:
        # 评估脚本的参数
        eval_params = {}
        if ALGORITHM_TYPE == 'classification':
            eval_params['image_shape'] = IMAGE_SHAPE
        elif ALGORITHM_TYPE == 'segmentation':
            eval_params['gpus'] = GPUS
            eval_params['val_data'] = VAL_DATA
            eval_params['eval_'] = 'true'
        # 从指定路径解析模型名和epoch
        model_prefix, epoch_list = _get_model_prefix_epoch(EVAL_MODEL)
        # 从评估脚本加载评估接口
        evaluate_function = _load_evaluate(EVAL_SCRIPT)

        if not epoch_list:
            logging.error('model path has no params')
            sys.exit(1)

        logging.info('input model prefix is %s' % model_prefix)
        logging.info(epoch_list)

        # 逐个对模型做评估,找到当前训练轮中最优的
        index = -1
        max_value = -1

        logging.info(CLASSES)
        _classes = b64decode(CLASSES)
        logging.info('eval classes: {}'.format(_classes))
        # 改为只取最后一个checkpoint
        epoch_list.sort()
        new_epoch_list = [epoch_list[-1]]
        for i in range(0, len(new_epoch_list)):
            # 返回归一化数值用于比较
            value = evaluate_function(model_prefix=model_prefix, \
                                      load_epoch=new_epoch_list[i],
                                      test_data=VAL_DATA,
                                      classes=_classes, **eval_params)
            logging.info('eval output')
            logging.info(value)
            if not isinstance(value, (int, float)):
                logging.error('evaluate script return value illegal')
                sys.exit(1)

            if value >= max_value:
                max_value = value
                index = i

        if index < 0:
            logging.error('there is no model file in the folder')
            sys.exit(1)

        logging.info('the best model epoch in this iter is %s' % new_epoch_list[index])
        # 如果模型比上次优，保存到系统
        status, new_model_prefix = _save_model(model_prefix=model_prefix,
                                               save_epoch=new_epoch_list[index],
                                               eval_value=max_value)
        if not status:
            sys.exit(1)

        logging.info('return model prefix is %s' % new_model_prefix)
        flag = 0 if new_model_prefix == model_prefix else 1
        # 将结果保存到临时文件，用于下游节点使用
        status, message = _save_result(new_model_prefix, flag, EVAL_OUTPUT_PATH)
        if not status:
            sys.exit(1)
    except Exception as err:
        logging.error('occured: {}'.format(str(err)))
        import traceback
        traceback.print_exc()
        sys.exit(1)


import os
import cv2
import logging
import traceback
import mxnet as mx
import multiprocessing
from aios.utils.utils import *

class SegmentationPack(object):

    def __init__(self, mount_path=None):
        self.mount_path = mount_path if not mount_path else '/datakubeflow'

    def create_segmentation_txt(self, save_path, labels):
        '''从DLP导出的json文件中提取原图和切割图，分别存入txt中
        '''
        prefix, img_fmt = os.path.splitext(save_path)
        img_txt = prefix + '_img.txt'
        mask_txt = prefix + '_mask.txt'

        logging.info('start get segmentation text...')
        imgs = []
        masks = []

        for it in labels["data"]:

            img_path = self.mount_path + '/' + it["host"] + '\n'
            if not it.get('matte_url'):
                continue
            mask_path = self.mount_path + '/' + it["matte_url"] + '\n'
            imgs.append(img_path)
            masks.append(mask_path)

        with open(os.path.join(img_txt), "w") as f:
            f.writelines(imgs)

        with open(os.path.join(mask_txt), "w") as f:
            f.writelines(masks)

class MultiProcessRecPack(object):
  
    def __init__(self, mount_path=None, image_shape=None):
        self.mount_path = mount_path if not mount_path else '/datakubeflow'
        self.type = None
        self.image_shape = image_shape
        
    def image_encode(self, i, item, q_out,  pass_through=False):
        # 将lst中的文件读出来，获取label以及图片路径，读取图片数据与header形成元组(i, s, item )压入队列

        # fullpath = item["url"]
        fullpath = item["host"]
        fullpath = self.mount_path + '/' + fullpath
        quality, encoding = get_img_fmt_quality(fullpath)

        if pass_through:
            # 二进制读取模式，使用该模式就不会用下面的cv2.imread读取方式
            # 这种模式不能做resize,center_crop等变换操作，只能原图是什么就是什么
            label = self.get_label(self.type, item)
            if not label:
                q_out.put((i, None, item))
                logging.error('label is null')
                return

            header = mx.recordio.IRHeader(0, label, item['id'], 0)
            try:
                with open(fullpath, 'rb') as fin:
                    img = fin.read()

                img = self.resize_img(img)

                s = mx.recordio.pack(header, img)
                q_out.put((i, s, item))
            except Exception as e:
                traceback.print_exc()
                print('pack_img error:', item["url"], e)
                q_out.put((i, None, item))
            return
        else:
            try:
                img = cv2.imread(fullpath, 1)  # color=1是彩色读取
            except:
                traceback.print_exc()
                print('imread error trying to load file: %s ' % fullpath)
                q_out.put((i, None, item))
                return
            if img is None:
                print('imread read blank (None) image for file: %s' % fullpath)
                q_out.put((i, None, item))
                return

            try:
                # img是已经读取到内存的，不管图片本身的格式，只要设定好quality和编码方式
                # 如果不做变换，跟pass_through的处理结果是一样的
                h, w, c = img.shape
                label = get_label(self.type, item, w, h)
                if label is None:
                    q_out.put((i, None, item))
                    logging.error('label is null')
                    return
                
                img = self.resize_img(img)

                header = mx.recordio.IRHeader(0, label, item['id'], 0)
                s = mx.recordio.pack_img(header, img, quality=quality, img_fmt=encoding)
                q_out.put((i, s, item))
            except Exception as e:
                traceback.print_exc()
                print('pack_img error on file: %s' % fullpath, e)
                q_out.put((i, None, item))
                return

    def read_worker(self, q_in, q_out):
        while True:
            # 从q_in队列中取出数据处理完之后存到q_out
            deq = q_in.get()
            if deq is None:
                break
            i, item = deq
            self.image_encode(i, item, q_out)

    def write_worker(self, q_out, fname):
        """
        :param q_out: 从q_out中读取数据打包
        :param fname:  rec文件名的全路径
        :return:
        """
        logging.info('write worker start...')
        count = 0
        fname_rec = fname
        fname_idx = os.path.splitext(fname)[0] + '.idx'

        logging.info('{}, {}'.format(fname_idx, fname_rec))
        record = mx.recordio.MXIndexedRecordIO(fname_idx, fname_rec, 'w')
        buf = {}
        more = True
        while more:
            deq = q_out.get()
            if deq is not None:
                i, s, item = deq
                buf[i] = (i, s, item)
            else:
                more = False

            while count in buf:
                i, s, item, = buf[count]
                del buf[count]
                if s is not None:
                    record.write_idx(i, s)  # 第一个参数是索引即lst文件中的第一列，如果索引错了随机读取会出错，顺序读取不会出错

                count += 1

    def mxnet_multiprocess_handle(self, save_path, labels, process_num=multiprocessing.cpu_count() * 10):
        """
        使用多进程打包生成rec和idx, 参考im2rec.py, 多个进程读取image,写到rec文件只有一个进程
        lst的第一列是给图片编号，idx的第一列是对应同样的编号，idx的第二列对应的这个编号的图片的大小，rec是存储具体图片信息的文件
        """
        self.type = labels['label_type']
        image_list = labels["data"]
        process_num = 1 if process_num < 1 else process_num
        if len(image_list) < 100:
            process_num = 1

        # 初始化进程与队列
        logging.info('start pack to rec file with process num %s' % process_num)
        print('start pack to rec file with process num %s' % process_num)
        q_in = [multiprocessing.Queue(1024) for i in range(process_num)]
        q_out = multiprocessing.Queue(1024)
        read_process = [multiprocessing.Process(target=self.read_worker, args=(q_in[i], q_out)) \
                        for i in range(process_num)]
        for p in read_process:
            p.start()

        write_process = multiprocessing.Process(target=self.write_worker, args=(q_out, save_path))
        write_process.start()

        # 初始化队列，将图片均匀的分散到队列里面
        for i, item in enumerate(image_list):
            # item是lst文件中的每一行
            q_in[i % len(q_in)].put((i, item))

        for q in q_in:
            q.put(None)
        for p in read_process:
            p.join()

        q_out.put(None)
        write_process.join()

    def resize_img(self, img):
        # 图片压缩,在训练过程中应保持一致
        if self.image_shape:
            [height, width] = self.image_shape.split(',')
            resize = max(int(height), int(width))
            if img.shape[0] > img.shape[1]:
                newsize = (resize, img.shape[0] * resize // img.shape[1])
            else:
                newsize = (img.shape[1] * resize // img.shape[0], resize)
            img = cv2.resize(img, newsize)
        return img



class ModelClass(object):

    def __init__(self, **kwargs):
        # default
        self.num_classes = 12
        # momentum for batch normlization
        self.bn_mom = 0.9
        # memory space size(MB) used in convolution, if xpu memory is oom, then you can try smaller vale, such as --workspace 256
        self.workspace = 512
        # true means using memonger to save momory, https://github.com/dmlc/mxnet-memonger
        self.memonger = False

        for k,v in kwargs.items():
            setattr(self, k, v)
